/********************************************************************************

	GnucNS - Network Simulator
    Copyright (C) 2003 John Marshall

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	For support, questions, comments, etc...
	E-Mail: swabby@c0re.net

********************************************************************************/

using System;
using System.Collections;


namespace GnucNS.Simulation
{
	/// <summary>
	/// Summary description for SimNetwork.
	/// </summary>
	public class SimNetwork
	{
		public SimSetup Settings;

		public uint NetworkAlive = 0;

		public Hashtable NodeTable = new Hashtable();
		int NextNodeID = 1;
		
		public Hashtable WebCacheTable = new Hashtable();
		int NextWebCacheID = 1;

		public Random RndGen = new Random(unchecked((int)DateTime.Now.Ticks)); 

		public ArrayList PendingTrafficIDs = new ArrayList();


		public SimNetwork(SimSetup setup)
		{
			Settings = setup;

			SpawnWebCaches( Settings.StartingWebCaches);
			SpawnNodes( Settings.StartingNodes);
		}

		public void RunSecond()
		{
			NetworkAlive++;

			// Node flux
			if(Settings.NetworkInflux > 0 && NodeTable.Count < Settings.NetworkCapacity)
				if( RndGen.Next(Settings.NetworkInflux) == 0)
					SpawnNodes(1);

			if(Settings.NetworkOutflux > 0)
				if( RndGen.Next(Settings.NetworkOutflux) == 0)
					RemoveNodes(1);

			while( NodeTable.Count > Settings.NetworkCapacity)
				RemoveNodes(1);


			// Run Nodes
			IDictionaryEnumerator nodeEnum = NodeTable.GetEnumerator();

			while( nodeEnum.MoveNext() )
			{
				SimNode currentNode = (SimNode) nodeEnum.Value;
				currentNode.RunSecond();
			}

			if(Settings.TrafficSim)
				TrafficSim();
		}

		public void SpawnNodes(int amount)
		{
			while( amount > 0 )
			{
				SimNode spawnNode = new SimNode(this, NextNodeID);

			
				// Set bandwidth
				int nodeBandwidth = RndGen.Next( Settings.NodeBandwidthMin, Settings.NodeBandwidthMax );
				spawnNode.BandwidthTotal = nodeBandwidth;
				spawnNode.BandwidthUp = nodeBandwidth / 2;
				spawnNode.BandwidthDown = nodeBandwidth / 2;

				// Set Leaf Capacity
				spawnNode.LeafCapacity = RndGen.Next( Settings.NodeLeafMin, Settings.NodeLeafMax );

				// Set if firewalled
				if( RndGen.Next(100) < Settings.NetworkFirewalled)
				{
					spawnNode.Firewalled = true;
					spawnNode.LeafCapacity = 0;
				}

				// Add node to table
				NodeTable[NextNodeID] = spawnNode;
				NextNodeID++;
				amount--;
			}
		}
	
		public void RemoveNodes(int amount)
		{
			while( amount > 0 && NodeTable.Count > 0)
			{
				int curIndex = 0;
				int rndIndex = RndGen.Next( NodeTable.Count - 1 );

				IDictionaryEnumerator nodeEnum = NodeTable.GetEnumerator();

				while( nodeEnum.MoveNext() )
				{
					if(curIndex == rndIndex)
					{
						((SimNode) nodeEnum.Value).Close();

						NodeTable.Remove( nodeEnum.Key );
						break;
					}

					curIndex++;
				}

				amount--;
			}
		}

		public void SpawnWebCaches(int amount)
		{
			while( amount > 0 )
			{
				WebCacheTable[NextWebCacheID] = new SimWebCache(this, NextWebCacheID);

				NextWebCacheID++;
				amount--;
			}
		}

		public void TrafficSim()
		{
			while( PendingTrafficIDs.Count > 0 )
			{
				ArrayList pumpNodes = PendingTrafficIDs;
				PendingTrafficIDs = new ArrayList();

				
				foreach( int nodeID in pumpNodes)
					((SimNode) NodeTable[nodeID]).PumpPackets();
			}
		}

	}
}
