/********************************************************************************

	GnucNS - Network Simulator
    Copyright (C) 2003 John Marshall

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	For support, questions, comments, etc...
	E-Mail: swabby@c0re.net

********************************************************************************/

using System;
using System.Collections;
using System.Drawing;

namespace GnucNS.Simulation
{
	/// <summary>
	/// Summary description for SimNode.
	/// </summary>
	public class SimNode
	{		
		protected SimNetwork TestNet;

		public int  NodeID = 0;
		
		public ArrayList Connections = new ArrayList();
		public ArrayList NodeCache   = new ArrayList();
		
		public uint NodeAlive = 0;
		public bool Firewalled = false;
		public int AttemptID; // Current connection attempt
		
		int TrianglesInt = 0;


		// Ultrapeer
		public bool ChildMode = false;
		public int  LeafCapacity = 0;
		public ArrayList Children	 = new ArrayList();

		double Mod = 0.90;

		// Traffic
		public int BandwidthTotal = 0;
		public int BandwidthUp = 0;
		public int BandwidthDown = 0;
		public int BandwidthKnown = 0;

		public int SecBytesLeftUp = 0;
		public int SecBytesLeftDown = 0;
		public int SecBytesDropped = 0;
		public int SecBytesCollided = 0;

		public Queue LogBytesUp = new Queue();
		public Queue LogBytesDown = new Queue();
		public Queue LogBytesDropped = new Queue();
		public Queue LogBytesCollided = new Queue();

		public int TotBytesUp = 0;
		public int TotBytesDown = 0;
		public int TotBytesDropped = 0;
		public int TotBytesCollided = 0;

		int PacketBuffSize = 20;
		public Queue PacketQueue = new Queue();
		Hashtable PacketGuids = new Hashtable();
		int CleanTableInt = 0;


		public SimNode(SimNetwork theNetwork, int nodeID)
		{
			TestNet = theNetwork;
			NodeID  = nodeID;
		}

		public void RunSecond()
		{
			NodeAlive++;

			WebCacheTasks();
			ConnectTasks();
			TrafficTasks();
		}

		private void WebCacheTasks()
		{
			
			// If no connections, and nothing in cache, request nodes from WebCache
			if( Connections.Count == 0 && NodeCache.Count == 0 && TestNet.WebCacheTable.Count > 0)
			{
				SimWebCache rndCache = GetRandomWebCache();
				
				if(rndCache != null)
					rndCache.DoRequest(this);
			}

			// Update with random WebCache each hour
			if( !Firewalled && !ChildMode )
				if(NodeAlive % 3600 == 0)
				{
					SimWebCache rndCache = GetRandomWebCache();
					
					if(rndCache != null)
						rndCache.DoUpdate(this);
				}
		}

		private void ConnectTasks()
		{
			AttemptID = 0;

			int connectMin = ChildMode ? TestNet.Settings.LeafModeConnects : TestNet.Settings.NodeConnectMin ;
			int connectMax = ChildMode ? TestNet.Settings.LeafModeConnects : TestNet.Settings.NodeConnectMax ;
			
			// Check if node needs connections
			if(Connections.Count < connectMin && NodeCache.Count > 0)
			{
				// Get last node added to cache
				int rndIndex = TestNet.RndGen.Next( NodeCache.Count - 1);
				SimNode attemptNode = (SimNode) TestNet.NodeTable[ NodeCache[rndIndex] ];
				
				if(attemptNode != null)
				{
					AttemptID = attemptNode.NodeID;
					
					if( !Connections.Contains(AttemptID) && !Children.Contains(AttemptID) && AttemptID != NodeID )
						if(attemptNode.RequestConnect(this) )
						{
							if(attemptNode.ChildMode)
								Children.Add(AttemptID);
							else
								Connections.Add(AttemptID);

							AttemptID = 0;
						}
				}

				NodeCache.RemoveAt(rndIndex);
			}

			// If too many connections disconnect last connected node
			if(Connections.Count > connectMax)
			{
				int lastNodeID = (int) Connections[Connections.Count - 1];
				SimNode lastNode = (SimNode) TestNet.NodeTable[lastNodeID];

				lastNode.NotifyDisconnect(this);
				Connections.Remove( lastNodeID );
			}

			if(Children.Count > LeafCapacity)
			{
				int lastChildID = (int) Children[Children.Count - 1];
				SimNode lastChild = (SimNode) TestNet.NodeTable[lastChildID];

				lastChild.NotifyDisconnect(this);
				Children.Remove( lastChildID );
			}


			// Avoid Triangles
			AvoidTriangles();
		}	

		private SimWebCache GetRandomWebCache()
		{
			int webCacheIndex = TestNet.RndGen.Next( TestNet.WebCacheTable.Count - 1);

			IDictionaryEnumerator cacheEnum = TestNet.WebCacheTable.GetEnumerator();
			int currentIndex = 0;

			while( cacheEnum.MoveNext() )
				if(currentIndex == webCacheIndex)
					return (SimWebCache) cacheEnum.Value;
				else
					currentIndex++;

			return null;
		}

		public void UpdateCache(int remoteNodeID)
		{
			// commented out for speed
			if( /*!NodeCache.Contains(remoteNodeID) &&*/ remoteNodeID != NodeID)
				NodeCache.Add(remoteNodeID);

			if(NodeCache.Count > 50)
				NodeCache.RemoveAt(0);
		}


		private void AvoidTriangles()
		{
			if( !TestNet.Settings.NodeAvoidTriangles || ChildMode)
				return;
			
			if( TrianglesInt < 60 )
			{
				TrianglesInt++;
				return;
			}

			TrianglesInt = 0;


			Connections.Reverse();

			// For Each node we're connected to
			foreach( int currentNodeID in Connections)
			{
				// Compare current connection, with connections of other connections
				SimNode currentNode = (SimNode) TestNet.NodeTable[currentNodeID];

				foreach( int otherNodeID in Connections)
					if(otherNodeID != currentNodeID)
					{
						SimNode otherNode = (SimNode) TestNet.NodeTable[otherNodeID];
						
						// Compare this->1 hop with other->2 hop
						foreach( int remoteNodeID in otherNode.Connections)
							if(remoteNodeID == currentNodeID)
							{
								currentNode.NotifyDisconnect(this);
								Connections.Remove( currentNodeID );
								return;
							}

						
						foreach( int remoteNodeID in currentNode.Connections)
						{
							// Compare this->2 hops with other->1 hop
							if( remoteNodeID == otherNodeID)
							{
								currentNode.NotifyDisconnect(this);
								Connections.Remove( currentNodeID );
								return;
							}

							// Compare this->2 hops wiht other->2 hops
							foreach( int otherRemoteNodeID in otherNode.Connections)
								if( remoteNodeID == otherRemoteNodeID && otherRemoteNodeID != NodeID)
								{
									currentNode.NotifyDisconnect(this);
									Connections.Remove( currentNodeID );
									return;
								}
						}
					}
			}

			Connections.Reverse();
		}

		public bool RequestConnect( SimNode requester )
		{
			// If firewalled, node cant be connected to
			if(Firewalled)
				return false;


			// X-Try header
			foreach( int nodeID in Connections)
				requester.UpdateCache(nodeID);
			

			if( TestNet.Settings.UltrapeerSim )
			{
				// X-Ultrapeer: false
				if( requester.ChildMode )
				{
					if( Children.Count > LeafCapacity * Mod)
						if( requester.LeafCapacity > LeafCapacity * Mod &&
							requester.BandwidthKnown > BandwidthKnown * Mod &&
							requester.NodeAlive > NodeAlive * Mod)
						
							// X-Ultrapeer-Needed: true
							if( requester.AskUpgrade(this) )
							{
								// Node upgraded
								return false;
							}

					if( Children.Count < LeafCapacity )
					{
						Children.Add( requester.NodeID );
						return true;
					}

					return false;
				}
				

				// X-Ultrapeer: true
				if( requester.LeafCapacity * ( 1 - Mod) < LeafCapacity - Children.Count ||
					requester.BandwidthKnown < BandwidthKnown * Mod || 
					requester.NodeAlive < NodeAlive * Mod)
				
					// X-Ultrapeer-Needed: false
					if( requester.AskDowngrade(this) )
					{
						Children.Add( requester.NodeID );
						return true;
					}
			}

			if(Connections.Count < TestNet.Settings.NodeConnectMax)
			{
				Connections.Add(requester.NodeID);
				return true;
			}

			return false;
		}

		int FreeCapacity(int remoteLeafMax)
		{	
			int freeSlots = LeafCapacity - Children.Count;

			return freeSlots * 100 / remoteLeafMax;
		}
		
		int RunningCapacity(int remoteLeafMax)
		{	
			return Children.Count * 100 / remoteLeafMax;
		}

		public bool AskDowngrade(SimNode asking)
		{	
			if( Children.Count > asking.LeafCapacity * Mod ||
				asking.BandwidthKnown < BandwidthKnown * Mod ||
				asking.NodeAlive < NodeAlive * Mod)
				return false;

			Close();
			ResetTraffic();
			ChildMode = true;

			return true;
		}


		public bool AskUpgrade(SimNode asking)
		{
			Close();
			ChildMode = false;

			return true;		
		}

		public int RunningCapacity()
		{	
			return Children.Count * 100 / LeafCapacity;
		}

		public void NotifyDisconnect( SimNode notifier )
		{
			// Remove us from them
			if(notifier.ChildMode)
				Children.Remove( notifier.NodeID );
			else
				Connections.Remove( notifier.NodeID ); 
		}

		// Called before being deconstructed or shutting node down
		public void Close()
		{
			foreach( int nodeID in Connections)
				((SimNode) TestNet.NodeTable[nodeID]).NotifyDisconnect(this);

			foreach( int nodeID in Children)
				((SimNode) TestNet.NodeTable[nodeID]).NotifyDisconnect(this);

			Connections.Clear();
			Children.Clear();
		}

		void TrafficTasks()
		{
			if(TestNet.Settings.TrafficSim && !ChildMode)
			{
				LogBytesUp.Enqueue(BandwidthUp - SecBytesLeftUp);
				TotBytesUp += BandwidthUp - SecBytesLeftUp;
				if(LogBytesUp.Count > 60)
					TotBytesUp -= (int) LogBytesUp.Dequeue();
				SecBytesLeftUp   = BandwidthUp;

				LogBytesDown.Enqueue(BandwidthDown - SecBytesLeftDown);
				TotBytesDown += BandwidthDown - SecBytesLeftDown;
				if(LogBytesDown.Count > 60)
					TotBytesDown -= (int) LogBytesDown.Dequeue();
				SecBytesLeftDown = BandwidthDown;

				LogBytesDropped.Enqueue(SecBytesDropped);
				TotBytesDropped += SecBytesDropped;
				if(LogBytesDropped.Count > 60)
					TotBytesDropped -= (int) LogBytesDropped.Dequeue();
				SecBytesDropped  = 0;

				LogBytesCollided.Enqueue(SecBytesCollided);				
				TotBytesCollided += SecBytesCollided;	
				if(LogBytesCollided.Count > 60)
					TotBytesCollided -= (int) LogBytesCollided.Dequeue();
				SecBytesCollided = 0;


				int bwKnown = (TotBytesUp / LogBytesUp.Count) + (TotBytesDown / LogBytesDown.Count);

				if(bwKnown > BandwidthKnown)
					BandwidthKnown = bwKnown;


				// Check if random query will be sent
				if(TestNet.Settings.NodeQueryFreq > 0)
					if( TestNet.RndGen.Next(TestNet.Settings.NodeQueryFreq) == 0 )
					{
						Packet outPacket = new Packet( TestNet.RndGen.Next(int.MaxValue), 40);

						PacketGuids[outPacket.Guid] = NodeAlive;
						PacketQueue.Enqueue( outPacket );

						if(PacketQueue.Count > PacketBuffSize)
							SecBytesDropped += ((Packet) PacketQueue.Dequeue()).Size;
					}

				if( PacketQueue.Count > 0)
					TestNet.PendingTrafficIDs.Add( NodeID );

				// Remove packets from hash table older than 30 secs
				if(CleanTableInt >= 30)
				{
					ArrayList removeKeys = new ArrayList();

					IDictionaryEnumerator enumPacket = PacketGuids.GetEnumerator();
					while( enumPacket.MoveNext() )
						if( NodeAlive - ((uint) enumPacket.Value) > 30)
							removeKeys.Add((int)enumPacket.Key);

					foreach(int key in removeKeys)
						PacketGuids.Remove(key);

					CleanTableInt = 0;
				}
				else
					CleanTableInt++;

			}

			if( !TestNet.Settings.TrafficSim)
				BandwidthKnown = BandwidthTotal;

		}

		public void ResetTraffic()
		{
			SecBytesLeftUp = 0;
			SecBytesLeftDown = 0;
			SecBytesDropped = 0;
			SecBytesCollided = 0;

			LogBytesUp.Clear();
			LogBytesDown.Clear();
			LogBytesDropped.Clear();
			LogBytesCollided.Clear();

			TotBytesUp = 0;
			TotBytesDown = 0;
			TotBytesDropped = 0;
			TotBytesCollided = 0;
		}

		public void PumpPackets()
		{
			// Go through packet queue, broadcast out packets
			while( SecBytesLeftUp > 0 && PacketQueue.Count > 0 )
			{
				Packet outPacket = (Packet) PacketQueue.Dequeue();

				foreach( int connID in Connections)
					if(connID != outPacket.OriginID)
					{
						SimNode conn = (SimNode) TestNet.NodeTable[connID];

						if(conn.SecBytesLeftDown > outPacket.Size && SecBytesLeftUp > 0)
						{
							SecBytesLeftUp -= outPacket.Size;
							conn.TransmitPacket(NodeID, outPacket);
						}
						else
							SecBytesDropped += outPacket.Size;
					}
			}
				
		}

		public void TransmitPacket(int originID, Packet inPacket)
		{
			inPacket.OriginID = originID;
			inPacket.Hops++;

			SecBytesLeftDown -= inPacket.Size;

			// Check for packet collision
			if( PacketGuids[inPacket.Guid] != null)
			{
				SecBytesCollided += inPacket.Size;
				return;
			}

			// Check packet hops
			if(inPacket.Hops < TestNet.Settings.NetworkTTL)
			{
				PacketGuids[inPacket.Guid] = NodeAlive;
				PacketQueue.Enqueue( inPacket );

				if(PacketQueue.Count > PacketBuffSize)
					SecBytesDropped += ((Packet) PacketQueue.Dequeue()).Size;

				TestNet.PendingTrafficIDs.Add( NodeID );
			}
		}	

	}

	public struct Packet
	{
		public int OriginID;

		public int Guid;
		public int Hops;
		public int Size;

		public Packet(int guid, int size)
		{
			OriginID = 0;

			Guid = guid;
			Hops = 0;
			Size = size;
		}
	}
}


// If node can handle 33% of remote nodes capacity request downgrade

// Always keep network starved for ultrapeers

 
 
